
import edu.supercom.csdl.*;

public class VariableCondition {
  Value val;
  boolean neq;

  public VariableCondition(Value v) {
    val = v;
    neq = false;
  }

  public VariableCondition(Value v, boolean n) {
    val = v;
    neq = n;
  }

  public Value getValue() {
    return val;
  }

  public boolean isNotEq() {
    return neq;
  }

  public boolean isEq() {
    return !neq;
  }

  public boolean consistentWith(VariableCondition vc, Variable var) {
    if (isNotEq() && vc.isEq() && vc.getValue().equals(val))
      return false;
    if (isEq() && vc.isNotEq() && vc.getValue().equals(val))
      return false;
    if (isEq() && vc.isEq() && !vc.getValue().equals(val))
      return false;
    if (isNotEq() && vc.isNotEq() && !vc.getValue().equals(val) &&
	(var.getDomain().size() <= 2))
      return false;
    return true;
  }

  public boolean implies(VariableCondition vc, Variable var) {
    if (isEq() && vc.isEq() && vc.getValue().equals(val))
      return true;
    if (isEq() && vc.isNotEq() && !vc.getValue().equals(val))
      return true;
    if (isNotEq() && vc.isEq() && !vc.getValue().equals(val) &&
	(var.getDomain().size() == 2))
      return true;
    return false;
  }

  // note: this does not check that the domain is large enough
  // to accommodate two different =/= conditions
  public boolean consistentWith(VariableCondition vc) {
    if (isNotEq() && vc.isEq() && vc.getValue().equals(val))
      return false;
    if (isEq() && vc.isNotEq() && vc.getValue().equals(val))
      return false;
    if (isEq() && vc.isEq() && !vc.getValue().equals(val))
      return false;
    return true;
  }

  // note: does not check the case =/= -> == in binary domains
  public boolean implies(VariableCondition vc) {
    if (isEq() && vc.isEq() && vc.getValue().equals(val))
      return true;
    if (isEq() && vc.isNotEq() && !vc.getValue().equals(val))
      return true;
    return false;
  }

  public VariableCondition getNegation() {
    return new VariableCondition(val, !neq);
  }

  public String toString() {
    if (neq) {
      return "=/=" + val.toString();
    }
    else {
      return "==" + val.toString();
    }
  }
}
