
import java.io.*;
import java.util.*;
import java.util.regex.*;
import edu.supercom.csdl.*;

class csdltodot {

  public static void findStates(Iterator vars, Vector states)
  {
    if (vars.hasNext()) {
      Variable var = (Variable)vars.next();
      Vector ps = new Vector();
      findStates(vars, ps);
      Iterator vals = var.getDomain().valueIterator();
      while (vals.hasNext()) {
	Value val = (Value)vals.next();
	for (int k = 0; k < ps.size(); k++) {
	  HashMap xs = new HashMap((HashMap)ps.get(k));
	  xs.put(var, var.getAssignement(val));
	  states.add(xs);
	}
      }
    }
    else {
      states.add(new HashMap());
    }
  }

  public static void writeState(Map s, PrintStream out, String sep)
  {
    Iterator sass = s.entrySet().iterator();
    while (sass.hasNext()) {
      java.util.Map.Entry sa =
	(java.util.Map.Entry)sass.next();
      Variable var = (Variable)sa.getKey();
      Assignement val = (Assignement)sa.getValue();
      out.print(var.getName() + "=" + val.getValue().getLabel());
      if (sass.hasNext()) {
	out.print(sep);
      }
    }
  }

  public static void writeComponentStateGraph
    (ComponentType ct, PrintStream out)
    throws IOException
  {
    Vector states = new Vector();
    findStates(ct.variableIterator(), states);
    // System.err.println("component type " + ct.getName() + " has "
    //		       + states.size() + " states");

    out.println("digraph " + ct.getName() + " {");
    out.println("  label=\"" + ct.getName() + "\"");
    out.println("  node [shape=box,width=0,height=0];");
    out.println("  overlap=false;");
    out.println("  sep=0.5;");
    out.println("  splines=true;");

    for (int k = 0; k < states.size(); k++) {
      out.print("  S" + k + " [label=\"");
      writeState((Map)states.get(k), out, "\\n");
      out.println("\"];");
    }

    Iterator rules = ct.ruleIterator();
    while (rules.hasNext()) {
      Rule r = (Rule)rules.next();
      Event trig = r.getTriggeringEvent();
      for (int k = 0; k < states.size(); k++)
	if (r.satisfied((Map)states.get(k),
			ComponentStateCondition.ALL_VARIABLE_ASSIGNED)) {
	  HashMap ns = new HashMap((Map)states.get(k));
	  r.getEffect().doModify(ns, ComponentStateModification.ASSIGN_ALL);
	  int j = states.indexOf(ns);
	  if (j < 0) {
	    System.err.println("error: applying rule " + r.toString() +
			       " in state " + states.get(k).toString() +
			       " leads to " + ns.toString() +
			       " not in state set");
	    System.exit(255);
	  }
	  out.print("  S" + k + " -> S" + j + " [label=\"");
	  if (ct.isSpontaneousPort(trig.getPort())) {
	    out.print(trig.getMessage().getLabel());
	  }
	  else {
	    out.print(trig.getPort().getName() + ":" +
		      trig.getMessage().getLabel());
	  }
	  Iterator oevs = r.outputEventIterator();
	  if (oevs.hasNext()) {
	    out.print(" -> ");
	    while (oevs.hasNext()) {
	      Event e = (Event)oevs.next();
	      out.print(e.getPort().getName() + ":" +
			e.getMessage().getLabel());
	      if (oevs.hasNext()) {
		out.print(",");
	      }
	    }
	  }
	  out.println("\"];");
	}
    }

    out.println("}");
  }


  public static void main(String[] argv) {
    if (argv.length < 1) {
      System.err.println("csdltodot <system file>");
      System.exit(1);
    }
    String system_file = argv[0];
    try {
      CompositeSystem sys = CompositeSystem.load(system_file);
      Vector ctypes = new Vector();
      Iterator comps = sys.componentIterator();
      while (comps.hasNext()) {
	Component comp = (Component)comps.next();
	if (!ctypes.contains(comp.getType()))
	  ctypes.add(comp.getType());
      }
      System.err.println(ctypes.size() + " component types");
      for (int k = 0; k < ctypes.size(); k++) {
	writeComponentStateGraph((ComponentType)ctypes.get(k), System.out);
      }
    }
    catch (Exception e) {
      System.err.println("caught " + e.getMessage());
      e.printStackTrace();
    }
  }

}
