
import java.io.*;
import java.util.*;
import java.util.regex.*;
import edu.supercom.csdl.*;

class csdltopddl {
  static boolean strict_non_cyclic = true;
  static boolean write_comments = false;
  static boolean debug = false;

  // some utilities

  // make a (var, var_cond) map from a conjunctive condition

  private static boolean getConjuncts
    (ComponentStateCondition cond, Map m, int depth)
  {
    // System.err.println("getConjuncts " + depth + ": " + cond.toString());
    if (cond.isAnd()) {
      boolean v = getConjuncts(cond.getAndLeftPart(), m, depth + 1);
      v = (v && getConjuncts(cond.getAndRightPart(), m, depth + 1));
      return v;
    }
    else if (cond.isNegation()) {
      ComponentStateCondition not_cond = cond.getNegation();
      if (not_cond.isAssignement()) {
	Assignement a = not_cond.getAssignement();
	m.put(a.getVariable(), new VariableCondition(a.getValue(), true));
	return true;
      }
      else {
	System.err.println("error: can't handle condition " + cond.toString()
			   + " = not(" + not_cond.toString() + ")");
	System.exit(1);
	return false;
      }
    }
    else if (cond.isAssignement()) {
      Assignement a = cond.getAssignement();
      m.put(a.getVariable(), new VariableCondition(a.getValue()));
      return true;
    }
    else if (cond.isConstant()) {
      boolean v = cond.getConstant();
      return v;
    }
    else {
      System.err.println("error: can't handle condition " + cond.toString());
      System.exit(1);
      return false;
    }
  }

  // inverse of above: make a CSC from a (var, varcond) map

  private static ComponentStateCondition makeCondition(Map m)
  {
    ComponentStateCondition c = ComponentStateCondition.build(true);
    Iterator i = m.entrySet().iterator();
    while (i.hasNext()) {
      Map.Entry e = (Map.Entry)i.next();
      Variable var = (Variable)e.getKey();
      VariableCondition vc = (VariableCondition)e.getValue();
      Assignement ass = var.getAssignement(vc.getValue());
      if (vc.isNotEq()) {
	c = ComponentStateCondition.and
	  (c, ComponentStateCondition.not
	   (ComponentStateCondition.build(ass)));
      }
      else {
	c = ComponentStateCondition.and(c, ComponentStateCondition.build(ass));
      }
    }
    return c;
  }

  // pretty-print a (var, varcond) or (var, assign) map

  private static void writeCondition(Map m, PrintStream out)
  {
    Iterator i = m.entrySet().iterator();
    if (i.hasNext()) {
      while (i.hasNext()) {
	Map.Entry e = (Map.Entry)i.next();
	Variable var = (Variable)e.getKey();
	VariableCondition vc = null;
	try {
	  vc = (VariableCondition)e.getValue();
	}
	catch (ClassCastException cce) {
	  Assignement ass = (Assignement)e.getValue();
	  vc = new VariableCondition(ass.getValue());
	}
	out.print(var.getName() + vc.toString());
	if (i.hasNext())
	  out.print(" and ");
      }
    }
    else {
      out.print("true");
    }
  }

  // check consistency of two conditions

  private static boolean consistent
    (ComponentStateCondition c1, ComponentStateCondition c2)
  {
    // System.err.println("in consistent");
    HashMap m1 = new HashMap();
    getConjuncts(c1, m1, 0);
    HashMap m2 = new HashMap();
    getConjuncts(c2, m2, 0);
    Iterator i = m1.entrySet().iterator();
    while (i.hasNext()) {
      Map.Entry e = (Map.Entry)i.next();
      Variable var = (Variable)e.getKey();
      VariableCondition vc1 = (VariableCondition)e.getValue();
      if (m2.containsKey(var)) {
	VariableCondition vc2 = (VariableCondition)m2.get(var);
	if (vc1.isEq() && vc2.isEq() &&
	    !vc1.getValue().equals(vc2.getValue()))
	  return false;
	if (vc1.isEq() && vc2.isNotEq() &&
	    vc1.getValue().equals(vc2.getValue()))
	  return false;
	if (vc1.isNotEq() && vc2.isEq() &&
	    vc1.getValue().equals(vc2.getValue()))
	  return false;
	if (vc1.isNotEq() && vc2.isNotEq() &&
	    !vc1.getValue().equals(vc2.getValue()) &&
	    (var.getDomain().size() <= 2))
	  return false;
      }
    }
    return true;
  }

  // check commutativity of two rules

  private static boolean commutative(Rule r1, Rule r2)
  {
    // System.err.println("in commutative");
    HashMap e1 = new HashMap();
    r1.getEffect().doModify(e1, ComponentStateModification.ASSIGN_ALL);
    HashMap p1 = new HashMap();
    getConjuncts(r1.getPrecondition(), p1, 0);
    HashMap e2 = new HashMap();
    r2.getEffect().doModify(e2, ComponentStateModification.ASSIGN_ALL);
    HashMap p2 = new HashMap();
    getConjuncts(r2.getPrecondition(), p2, 0);
    Iterator i = e1.keySet().iterator();
    while (i.hasNext()) {
      Variable var = (Variable)i.next();
      if (e2.containsKey(var)) {
	// rules assign different values to a variable (var)
	if (!e1.get(var).equals(e2.get(var)))
	  return false;
      }
      // rule 1 assigns a variable in the precondition of rule 2
      if (p2.containsKey(var))
	return false;
    }
    Iterator j = e2.keySet().iterator();
    while (j.hasNext()) {
      Variable var = (Variable)j.next();
      // rule 2 assigns a variable in the precondition of rule 1
      if (p1.containsKey(var))
	return false;
    }
    return true;
  }


  // the lifted model

  private static void writeForwardAction
    (ComponentType sctype,
     Port sp,
     ComponentType rctype,
     Port rp,
     Message msg,
     PrintStream out)
  {
    // neither output nor input port is observable
    out.println("  (:action forward_" + sctype.getName() + "_" +
		       sp.getName() + "_" + rctype.getName() + "_" +
		       rp.getName() + "_" + msg.getLabel());
    out.println("    :parameters (?cout - component_" +
		       sctype.getName() + " ?cin - component_" +
		       rctype.getName() + ")");
    out.println("    :precondition (and" +
		       " (connected_" + sctype.getName() + "_" +
		       sp.getName() + "_" + rctype.getName() + "_" +
		       rp.getName() + " ?cout ?cin)" +
		       " (not (observable_" + sctype.getName() + "_" +
		       sp.getName() + " ?cout))" +
		       " (not (observable_" + rctype.getName() + "_" +
		       rp.getName() + " ?cin))" +
		       " (committed_" + sctype.getName() + "_" +
		       sp.getName() + "_" + msg.getLabel() + " ?cout))");
    out.println("    :effect (and (not (committed_" +
		       sctype.getName() + "_" + sp.getName() + "_" +
		       msg.getLabel() + " ?cout)) (committed_" +
		       rctype.getName() + "_" + rp.getName() + "_" +
		       msg.getLabel() + " ?in))");
    out.println("   )");

    // output port is observable
    out.println("  (:action forward_" + sctype.getName() + "_" +
		       sp.getName() + "_" + rctype.getName() + "_" +
		       rp.getName() + "_" + msg.getLabel() +
		       "_observe_out_port");
    out.println("    :parameters (?cout - component_" +
		       sctype.getName() + " ?cin - component_" +
		       rctype.getName() + " ?obs - observation)");
    out.println("    :precondition (and" +
		       " (connected_" + sctype.getName() + "_" +
		       sp.getName() + "_" + rctype.getName() + "_" +
		       rp.getName() + " ?cout ?cin)" +
		       " (observable_" + sctype.getName() + "_" +
		       sp.getName() + " ?cout)" +
		       " (not (observable_" + rctype.getName() + "_" +
		       rp.getName() + " ?cin))" +
		       " (committed_" + sctype.getName() + "_" +
		       sp.getName() + "_" + msg.getLabel() + " ?cout)" +
		       " (observation_of_" + sctype.getName() + "_" +
		       sp.getName() + "_" + msg.getLabel()
		       + " ?cout ?obs)" +
		       " (not (observed ?obs)) (possible ?obs))");
    out.println("    :effect (and (not (committed_" +
		       sctype.getName() + "_" + sp.getName() + "_" +
		       msg.getLabel() + " ?cout)) (committed_" +
		       rctype.getName() + "_" + rp.getName() + "_" +
		       msg.getLabel() + " ?in) (observed ?obs))");
    out.println("   )");

    // input port is observable
    out.println("  (:action forward_" + sctype.getName() + "_" +
		       sp.getName() + "_" + rctype.getName() + "_" +
		       rp.getName() + "_" + msg.getLabel() +
		       "_observe_in_port");
    out.println("    :parameters (?cout - component_" +
		       sctype.getName() + " ?cin - component_" +
		       rctype.getName() + " ?obs - observation)");
    out.println("    :precondition (and" +
		       " (connected_" + sctype.getName() + "_" +
		       sp.getName() + "_" + rctype.getName() + "_" +
		       rp.getName() + " ?cout ?cin)" +
		       " (not (observable_" + sctype.getName() + "_" +
		       sp.getName() + " ?cout))" +
		       " (observable_" + rctype.getName() + "_" +
		       rp.getName() + " ?cin)" +
		       " (committed_" + sctype.getName() + "_" +
		       sp.getName() + "_" + msg.getLabel() + " ?cout)" +
		       " (observation_of_" + rctype.getName() + "_" +
		       rp.getName() + "_" + msg.getLabel()
		       + " ?cin ?obs)" +
		       " (not (observed ?obs)) (possible ?obs))");
    out.println("    :effect (and (not (committed_" +
		       sctype.getName() + "_" + sp.getName() + "_" +
		       msg.getLabel() + " ?cout)) (committed_" +
		       rctype.getName() + "_" + rp.getName() + "_" +
		       msg.getLabel() + " ?in) (observed ?obs))");
    out.println("   )");
  }

  private static void writeDiscardAction
    (ComponentType sctype, Port sp, Message msg, PrintStream out)
  {
    // non-observable
    out.println("  (:action discard_" + sctype.getName() + "_" +
		       sp.getName() + "_" + msg.getLabel());
    out.println("    :parameters (?cout - component_" +
		       sctype.getName() + ")");
    out.println("    :precondition (and" +
		       " (disconnected_" + sctype.getName() + "_" +
		       sp.getName() + " ?cout)" +
		       " (committed_" + sctype.getName() + "_" +
		       sp.getName() + "_" + msg.getLabel() + " ?cout))");
    out.println("    :effect (not (committed_" +
		       sctype.getName() + "_" + sp.getName() + "_" +
		       msg.getLabel() + " ?cout))");
    out.println("   )");

    // observable
    out.println("  (:action discard_" + sctype.getName() + "_" +
		       sp.getName() + "_" + msg.getLabel() + "_observe");
    out.println("    :parameters (?cout - component_" +
		       sctype.getName() + " ?obs - observation)");
    out.println("    :precondition (and" +
		       " (disconnected_" + sctype.getName() + "_" +
		       sp.getName() + " ?cout)" +
		       " (committed_" + sctype.getName() + "_" +
		       sp.getName() + "_" + msg.getLabel() + " ?cout)" +
		       " (observation_of_" + sctype.getName() + "_" +
		       sp.getName() + "_" + msg.getLabel()
		       + " ?cout ?obs)" +
		       " (not (observed ?obs)) (possible ?obs))");
    out.println("    :effect (not (committed_" +
		       sctype.getName() + "_" + sp.getName() + "_" +
		       msg.getLabel() + " ?cout) (observed ?obs))");
    out.println("   )");
  }

  private static void writeDomain(CompositeSystem sys, PrintStream out)
  {
    out.println("(define (domain system)");

    // type declarations
    out.print("  (:types observation component - object");
    Iterator ctypes = sys.typeIterator();
    while (ctypes.hasNext()) {
      out.println();
      ComponentType ctype = (ComponentType)ctypes.next();
      out.print("          component_"
		       + ctype.getName() +
		       " - component");
    }
    out.println(")");

    // predicate declarations
    out.println("  (:predicates");
    ctypes = sys.typeIterator();
    while (ctypes.hasNext()) {
      ComponentType ctype = (ComponentType)ctypes.next();
      Iterator cvars = ctype.variableIterator();
      while (cvars.hasNext()) {
	Variable cvar = (Variable)cvars.next();
	Domain cdom = cvar.getDomain();
	Iterator dvals = cdom.valueIterator();
	while (dvals.hasNext()) {
	  Value dval = (Value)dvals.next();
	  out.println("    (state_" + ctype.getName() +
			     "_" + cvar.getName() + 
			     "_" + dval.getLabel() +
			     " ?c - component_" +
			     ctype.getName() + ")");
	}
      }
      Iterator cinps = ctype.inputPortIterator();
      while (cinps.hasNext()) {
	Port cinp = (Port)cinps.next();
	if (!ctype.isSpontaneousPort(cinp)) {
	  Iterator pmsgs = cinp.messageIterator();
	  while (pmsgs.hasNext()) {
	    Message pmsg = (Message)pmsgs.next();
	    out.println("    (commmitted_" +
			       ctype.getName() +
			       "_" + cinp.getName() +
			       "_" + pmsg.getLabel() +
			       " ?c - component_" +
			       ctype.getName() + ")");
	  }
	}
      }
      Iterator coutps = ctype.outputPortIterator();
      while (coutps.hasNext()) {
	Port coutp = (Port)coutps.next();
	Iterator pmsgs = coutp.messageIterator();
	while (pmsgs.hasNext()) {
	  Message pmsg = (Message)pmsgs.next();
	  out.println("    (commmitted_" +
			     ctype.getName() +
			     "_" + coutp.getName() +
			     "_" + pmsg.getLabel() +
			     " ?c - component_" +
			     ctype.getName() + ")");
	}
      }
    }
    Iterator sctypes = sys.typeIterator();
    while (sctypes.hasNext()) {
      ComponentType sctype = (ComponentType)sctypes.next();
      Iterator sports = sctype.outputPortIterator();
      while (sports.hasNext()) {
	Port sp = (Port)sports.next();
	Iterator rctypes = sys.typeIterator();
	while (rctypes.hasNext()) {
	  ComponentType rctype = (ComponentType)rctypes.next();
	  Iterator rports = rctype.inputPortIterator();
	  while (rports.hasNext()) {
	    Port rp = (Port)rports.next();
	    if (sp.connectable(rp)) {
	      out.println("    (connected_" +
				 sctype.getName() + "_" +
				 sp.getName() + "_" +
				 rctype.getName() + "_" +
				 rp.getName() +
				 " ?cout - component_" +
				 sctype.getName() +
				 " ?cin - component_" +
				 rctype.getName() + ")");
	    }
	  }
	}
	out.println("    (disconnected_" + sctype.getName() +
			   "_" + sp.getName() + " ?cout - component_" +
			   sctype.getName() + ")");
      }
    }

    ctypes = sys.typeIterator();
    while (ctypes.hasNext()) {
      ComponentType ctype = (ComponentType)ctypes.next();
      Iterator ports = ctype.portIterator();
      while (ports.hasNext()) {
	Port p = (Port)ports.next();
	Iterator msgs = p.messageIterator();
	while (msgs.hasNext()) {
	  Message m = (Message)msgs.next();
	  out.println("    (observation_of_" + ctype.getName()
			     + "_" + p.getName() + "_" + m.getLabel()
			     + " ?c - component_" + ctype.getName()
			     + " ?o - observation)");
	}
      }
    }

    out.println("    (precedes ?o1 - observation ?o2 - observation)");
    out.println("    (possible ?o - observation)");
    out.println("    (observed ?o - observation)");
    out.println("    (idle))");

    // component rule actions
    ctypes = sys.typeIterator();
    while (ctypes.hasNext()) {
      ComponentType ctype = (ComponentType)ctypes.next();
      Iterator crules = ctype.ruleIterator();
      int rindex = 1;
      while (crules.hasNext()) {
	Rule crule = (Rule)crules.next();
	Port rport = crule.getTriggeringEvent().getPort();
	Message rmsg = crule.getTriggeringEvent().getMessage();
	out.println("  (:action rule_" + ctype.getName() +
			   "_" + rport.getName() +
			   "_" + rmsg.getLabel() +
			   "_" + Integer.toString(rindex));
	out.println("    :parameters (?c - component_" +
			   ctype.getName() + ")");
	out.print("    :precondition (and");
	if (!ctype.isSpontaneousPort(rport)) {
	  out.print(" (committed_" + ctype.getName() +
			   "_" + rport.getName() +
			   "_" + rmsg.getLabel() + " ?c)");
	}
	else {
	  out.print(" (idle)");
	}
	HashMap prec = new HashMap();
	getConjuncts(crule.getPrecondition(), prec, 0);
	Iterator precs = prec.entrySet().iterator();
	while (precs.hasNext()) {
	  java.util.Map.Entry pa =
	    (java.util.Map.Entry)precs.next();
	  Variable pavar = (Variable)pa.getKey();
	  Value paval = (Value)pa.getValue();
	  out.print(" (state_" + ctype.getName() +
			   "_" + pavar.getName() +
			   "_" + paval.getLabel() +
			   " ?c)");
	}
	out.println(")");
	out.print("    :effect (and");
	if (!ctype.isSpontaneousPort(rport)) {
	  out.print(" (not (committed_" + ctype.getName()
			   + "_" + rport.getName() + "_"
			   + rmsg.getLabel() + " ?c))");
	}
	Iterator routs = crule.outputEventIterator();
	while (routs.hasNext()) {
	  Event routev = (Event)routs.next();
	  out.print(" (committed_" + ctype.getName() +
			   "_" + routev.getPort().getName() +
			   "_" + routev.getMessage().getLabel()
			   + " ?c)");
	}

	HashMap rmap = new HashMap();
	crule.getEffect().doModify(rmap, ComponentStateModification.ASSIGN_ALL);
	Iterator ras = rmap.entrySet().iterator();
	while (ras.hasNext()) {
	  java.util.Map.Entry ra =
	    (java.util.Map.Entry)ras.next();
	  Variable ravar = (Variable)ra.getKey();
	  Value raval = ((Assignement)ra.getValue()).getValue();
	  if (prec.containsKey(ravar)) {
	    Value paval = (Value)prec.get(ravar);
	    if (!paval.equals(raval)) {
	      out.print(" (not (state_" +
			       ctype.getName() + "_" +
			       ravar.getName() + "_" +
			       paval.getLabel() + " ?c))");
	    }
	  }
	  else {
	    // ah, fuck it... not going to happen anyway
	    System.exit(255);
	  }
	  out.print(" (state_" +
			   ctype.getName() + "_" +
			   ravar.getName() + "_" +
			   raval.getLabel() + " ?c)");
	}
	out.println(")");

	// end of action
	out.println("   )");
	rindex += 1;
      }
    }

    // message propagation actions
    sctypes = sys.typeIterator();
    while (sctypes.hasNext()) {
      ComponentType sctype = (ComponentType)sctypes.next();
      Iterator sports = sctype.outputPortIterator();
      while (sports.hasNext()) {
	Port sp = (Port)sports.next();
	Iterator rctypes = sys.typeIterator();
	while (rctypes.hasNext()) {
	  ComponentType rctype = (ComponentType)rctypes.next();
	  Iterator rports = rctype.inputPortIterator();
	  while (rports.hasNext()) {
	    Port rp = (Port)rports.next();
	    if (sp.connectable(rp)) {
	      Iterator msgs = sp.messageIterator();
	      while (msgs.hasNext()) {
		Message msg = (Message)msgs.next();
		if (rp.getMessage(msg.getLabel()) != null) {
		  writeForwardAction(sctype, sp, rctype, rp, msg, out);
		}
	      }
	    }
	  }
	}
	Iterator msgs = sp.messageIterator();
	while (msgs.hasNext()) {
	  Message msg = (Message)msgs.next();
	  writeDiscardAction(sctype, sp, msg, out);
	}
      }
    }

    // observation update action
    out.println("  (:action next_observation");
    out.println("    :parameters (?o - observation)");
    out.println("    :precondition (forall (?op - observation)");
    out.println("                    (imply (precedes ?op ?p)");
    out.println("                      (observed ?op)))");
    out.println("    :effect (possible ?o)");
    out.println("   )");

    // definition of "idle"
    out.println("  (:derived (idle)");
    out.print("    (and");
    ctypes = sys.typeIterator();
    while (ctypes.hasNext()) {
      ComponentType ctype = (ComponentType)ctypes.next();
      Iterator cinps = ctype.inputPortIterator();
      while (cinps.hasNext()) {
	Port cinp = (Port)cinps.next();
	if (!ctype.isSpontaneousPort(cinp)) {
	  Iterator pmsgs = cinp.messageIterator();
	  while (pmsgs.hasNext()) {
	    Message pmsg = (Message)pmsgs.next();
	    out.print(" (forall (?c - component_"
			     + ctype.getName()
			     + ") (not (commmitted_" +
			     ctype.getName() +
			     "_" + cinp.getName() +
			     "_" + pmsg.getLabel() +
			     " ?c)))");
	  }
	}
      }
      Iterator coutps = ctype.outputPortIterator();
      while (coutps.hasNext()) {
	Port coutp = (Port)coutps.next();
	Iterator pmsgs = coutp.messageIterator();
	while (pmsgs.hasNext()) {
	  Message pmsg = (Message)pmsgs.next();
	  out.print(" (forall (?c - component_"
			   + ctype.getName()
			   + ") (not (commmitted_" +
			   ctype.getName() +
			   "_" + coutp.getName() +
			   "_" + pmsg.getLabel() +
			   " ?c)))");
	}
      }
    }
    out.println("))");

    // end of domain definition
    out.println("  )");
  }


  // propositional model

  public static void printEventCascade(Map casc, PrintStream out)
  {
    Iterator crs = casc.entrySet().iterator();
    while (crs.hasNext()) {
      Map.Entry e = (Map.Entry)crs.next();
      Component c = (Component)e.getKey();
      Iterator rs = ((Set)e.getValue()).iterator();
      while (rs.hasNext()) {
	Rule r = (Rule)rs.next();
	out.println("-> " + c.getName()
		    + " " + r.getTriggeringEvent().getPort().getName()
		    + ":" + r.getTriggeringEvent().getMessage().getLabel()
		    + " / " + r.toString());
      }
    }
  }

  private static Set extendCascades
    (CompositeSystem sys,
     Map ccrules,
     Component comp,
     Set cset,
     Iterator pending)
  {
    if (debug) {
      System.err.println("extendCascades: " + comp.getName());
    }
    if (pending.hasNext()) {
      Event e = (Event)pending.next();
      Port p = e.getPort();
      if (sys.getInitialTopology().isConnected(comp, p)) {
	Iterator conns =
	  sys.getInitialTopology().connectionIterator(comp, p);
	Message next_msg = e.getMessage();
	while (conns.hasNext()) {
	  Connection conn = (Connection)conns.next();
	  Component next_comp = conn.getTargetComponent();
	  Port next_port = conn.getTargetPort();
	  HashSet ext = new HashSet();
	  Iterator curs = cset.iterator();
	  while (curs.hasNext()) {
	    HashMap cur = (HashMap)curs.next();
	    Set s =
	      findCascades(sys, ccrules, next_comp, next_port, next_msg, cur);
	    ext.addAll(s);
	  }
	  cset = ext;
	}
      }
      return extendCascades(sys, ccrules, comp, cset, pending);
    }
    return cset;
  }

  private static Set findCascades
    (CompositeSystem sys,
     Map ccrules,
     Component comp,
     Port port,
     Message msg,
     HashMap cur)
  {
    if (debug) {
      System.err.println("findCascades: " + comp.getName() + " " + port.getName() + msg.getLabel());
    }
    HashSet cset = new HashSet();
    if (cur.containsKey(comp) && strict_non_cyclic) {
      System.err.println("error: cyclic event cascade detected");
      printEventCascade(cur, System.err);
      System.err.println("-> " + comp.getName()
			 + " " + port.getName() + ":" + msg.getLabel());
      System.exit(255);
    }
    Iterator rules = comp.ruleIterator();
    if (ccrules.containsKey(comp.getType())) {
      Iterator more_rules =
	((Collection)ccrules.get(comp.getType())).iterator();
      rules = new ChainIterator(rules, more_rules);
    }
    while (rules.hasNext()) {
      Rule rule = (Rule)rules.next();
      if (rule.getTriggeringEvent().getPort().equals(port) &&
	  rule.getTriggeringEvent().getMessage().equals(msg)) {
	HashMap new_cur = (HashMap)cur.clone();
	if (new_cur.containsKey(comp)) {
	  boolean firable = true;
	  Set prev_set = (Set)new_cur.get(comp);
	  Iterator rs = prev_set.iterator();
	  while (rs.hasNext() && firable) {
	    Rule r = (Rule)rs.next();
	    if (consistent(r.getPrecondition(), rule.getPrecondition())) {
	      if (r.getTriggeringEvent().getPort().equals(port) &&
		  r.getTriggeringEvent().getMessage().equals(msg)) {
		System.err.println("error: triggering cycle detected");
		printEventCascade(cur, System.err);
		System.err.println("-> " + comp.getName()
				   + " " + port.getName() + ":"
				   + msg.getLabel());
		System.exit(255);
	      }
	      if (!commutative(r, rule)) {
		System.err.println("error: non-commutative cyclic cascade detected");
		printEventCascade(cur, System.err);
		System.err.println("-> " + comp.getName()
				   + " " + port.getName() + ":"
				   + msg.getLabel()
				   + " / " + r.toString() 
				   + " and " + rule.toString());
		System.exit(255);
	      }
	    }
	    else {
	      firable = false;
	    }
	  }
	  if (firable || debug) {
	    HashSet new_set = new HashSet(prev_set);
	    new_set.add(rule);
	    new_cur.put(comp, new_set);
	  }
	}
	else {
	  HashSet new_set = new HashSet();
	  new_set.add(rule);
	  new_cur.put(comp, new_set);
	}
	HashSet new_cset = new HashSet();
	new_cset.add(new_cur);
	Set ext_cset = extendCascades(sys, ccrules, comp, new_cset,
				      rule.outputEventIterator());
	cset.addAll(ext_cset);
      }
    }
    return cset;
  }

  private static void writePropositionalAction
    (CompositeSystem sys,
     Component comp,
     Port port,
     Message msg,
     int index,
     HashMap act,
     PrintStream out)
  {
    if (write_comments) {
      out.println();
      Iterator crs = act.entrySet().iterator();
      while (crs.hasNext()) {
	Map.Entry e = (Map.Entry)crs.next();
	Component c = (Component)e.getKey();
	Set rset = (Set)e.getValue();
	Iterator rs = rset.iterator();
	while (rs.hasNext()) {
	  Rule r = (Rule)rs.next();
	  out.print("  ;; " + c.getName() + ": "
		    + r.getTriggeringEvent().getPort().getName()
		    + "." + r.getTriggeringEvent().getMessage().getLabel()
		    + " / ");
	  HashMap m = new HashMap();
	  getConjuncts(r.getPrecondition(), m, 0);
	  writeCondition(m, out);
	  out.print(" -> {");
	  Iterator roes = r.outputEventIterator();
	  while (roes.hasNext()) {
	    Event roe = (Event)roes.next();
	    out.print(roe.getPort().getName() + "."
		      + roe.getMessage().getLabel());
	    if (roes.hasNext())
	      out.print(", ");
	  }
	  out.print("} / ");
	  m = new HashMap();
	  r.getEffect().doModify(m, ComponentStateModification.ASSIGN_ALL);
	  writeCondition(m, out);
	  out.println();
	}
      }
    }

    out.println("  (:action " + comp.getName()
		+ "_" + port.getName() + "_" + msg.getLabel()
		+ "_" + Integer.toString(index));
    // collect observations and event occurrance sets
    HashSet obs = new HashSet();
    HashSet occ = new HashSet();
    Iterator crs = act.entrySet().iterator();
    while (crs.hasNext()) {
      Map.Entry e = (Map.Entry)crs.next();
      Component c = (Component)e.getKey();
      Set rset = (Set)e.getValue();
      Iterator rs = rset.iterator();
      while (rs.hasNext()) {
	Rule r = (Rule)rs.next();
	ComponentEvent ce = c.getComponentEvent(r.getTriggeringEvent());
	occ.add(ce);
	if (sys.getInitialTopology().isObservable(c, ce.getEvent().getPort()))
	  obs.add(ce);
	Iterator roes = r.outputEventIterator();
	while (roes.hasNext()) {
	  Event roe = (Event)roes.next();
	  ce = c.getComponentEvent(roe);
	  occ.add(ce);
	  if (sys.getInitialTopology().isObservable(c, ce.getEvent().getPort()))
	    obs.add(ce);
	}
      }
    }

    // parameters: one for each observable event
    out.print("    :parameters (");
    for (int i = 1; i <= obs.size(); i++) {
      if (i > 1) out.print(" ");
      out.print("?o" + Integer.toString(i));
    }
    out.println(")");

    // preconditions:
    out.print("    :precondition (and");
    // 1. state conditions required by rules
    crs = act.entrySet().iterator();
    while (crs.hasNext()) {
      Map.Entry e = (Map.Entry)crs.next();
      Component c = (Component)e.getKey();
      Set rset = (Set)e.getValue();
      Iterator rs = rset.iterator();
      while (rs.hasNext()) {
	Rule r = (Rule)rs.next();
	HashMap rc = new HashMap();
	// System.err.println("getConjuncts: " + c.toString());
	boolean is_sat = getConjuncts(r.getPrecondition(), rc, 0);
	Iterator rcs = rc.entrySet().iterator();
	while (rcs.hasNext()) {
	  Map.Entry rca = (Map.Entry)rcs.next();
	  Variable rcvar = (Variable)rca.getKey();
	  VariableCondition rcvc = (VariableCondition)rca.getValue();
	  if (rcvc.isNotEq()) {
	    out.print(" (not (state_" + c.getName() + "_" + rcvar.getName()
		      + "_" + rcvc.getValue().getLabel() + "))");
	  }
	  else {
	    out.print(" (state_" + c.getName() + "_" + rcvar.getName()
		      + "_" + rcvc.getValue().getLabel() + ")");
	  }
	}
      }
    }
    // 2. match and pending on each observation parameter
    Iterator obss = obs.iterator();
    int counter = 1;
    while (obss.hasNext()) {
      ComponentEvent ce = (ComponentEvent)obss.next();
      out.print(" (matches_" + ce.getComponent().getName()
		       + "_" + ce.getEvent().getPort().getName()
		       + "_" + ce.getEvent().getMessage().getLabel()
		       + " ?o" + Integer.toString(counter) + ")");
      out.print(" (pending ?o" + Integer.toString(counter++) + ")");
    }    
    out.println(")");

    // effects:
    out.print("    :effect (and");
    // 1. reassignments of component state variables
    crs = act.entrySet().iterator();
    while (crs.hasNext()) {
      Map.Entry e = (Map.Entry)crs.next();
      Component c = (Component)e.getKey();
      Set rset = (Set)e.getValue();
      Iterator rs = rset.iterator();
      while (rs.hasNext()) {
	Rule r = (Rule)rs.next();
	HashMap rc = new HashMap();
	// System.err.println("getConjuncts: " + c.toString());
	boolean is_sat = getConjuncts(r.getPrecondition(), rc, 0);
	HashMap re = new HashMap();
	r.getEffect().doModify(re, ComponentStateModification.ASSIGN_ALL);
	Iterator res = re.entrySet().iterator();
	while (res.hasNext()) {
	  Map.Entry rea = (Map.Entry)res.next();
	  Variable revar = (Variable)rea.getKey();
	  Assignement reass = (Assignement)rea.getValue();
	  if (rc.containsKey(revar)) {
	    VariableCondition rcvc = (VariableCondition)rc.get(revar);
	    if (!(rcvc.isEq() && reass.getValue().equals(rcvc.getValue()))) {
	      out.print(" (not (state_" + c.getName()
			+ "_" + revar.getName()
			+ "_" + rcvc.getValue().getLabel() + "))");
	      out.print(" (state_" + c.getName()
			+ "_" + revar.getName()
			+ "_" + reass.getValue().getLabel() + ")");
	    }
	  }
	  else {
	    Iterator vals = revar.getDomain().valueIterator();
	    while (vals.hasNext()) {
	      Value val = (Value)vals.next();
	      if (!val.equals(reass.getValue())) {
		out.print(" (not (state_" + c.getName()
			  + "_" + revar.getName()
			  + "_" + val.getLabel() + "))");
	      }
	    }
	    out.print(" (state_" + c.getName()
		      + "_" + revar.getName()
		      + "_" + reass.getValue().getLabel() + ")");
	  }
	}
      }
    }
    // 2. marking of event occurrences
    Iterator occs = occ.iterator();
    while (occs.hasNext()) {
      ComponentEvent ce = (ComponentEvent)occs.next();
      out.print(" (occurred_" + ce.getComponent().getName()
		       + "_" + ce.getEvent().getPort().getName()
		       + "_" + ce.getEvent().getMessage().getLabel() + ")");
    }    
    // 3. changing status of observations
    for (int i = 1; i <= obs.size(); i++) {
      out.print(" (not (pending ?o" + Integer.toString(i) + "))");
      out.print(" (observed ?o" + Integer.toString(i) + ")");
    }

    // end effects and end action
    out.println("))");
  }

  private static void writeSystemPropositional
    (CompositeSystem sys, Map ccrules, PrintStream out)
  {
    out.println("(define (domain SYSTEM)");

    // predicate declarations
    out.println("  (:predicates");
    Iterator comps = sys.componentIterator();
    while (comps.hasNext()) {
      // atoms for component state
      Component comp = (Component)comps.next();
      Iterator vars = comp.variableIterator();
      while (vars.hasNext()) {
	Variable var = (Variable)vars.next();
	Iterator vals = var.getDomain().valueIterator();
	while (vals.hasNext()) {
	  Value val = (Value)vals.next();
	  out.println("    (state_" + comp.getName()
			     + "_" + var.getName() + "_" + val.getLabel()
			     + ")");
	}
      }
      // atoms for event occurrences/observation matching
      Iterator events = comp.getType().eventIterator();
      while (events.hasNext()) {
	Event event = (Event)events.next();
	Port port = event.getPort();
	Message msg = event.getMessage();
	out.println("    (occurred_" + comp.getName()
			   + "_" + port.getName() + "_" + msg.getLabel()
			   + ")");
	if (sys.getInitialTopology().isObservable(comp, port)) {
	  out.println("    (matches_" + comp.getName()
			     + "_" + port.getName() + "_" + msg.getLabel()
			     + " ?obs)");
	}
      }
    }
    out.println("    (precedes ?obs1 ?obs2)");
    out.println("    (future ?obs)");
    out.println("    (pending ?obs)");
    out.println("    (observed ?obs))");

    // system rule actions
    comps = sys.componentIterator();
    while (comps.hasNext()) {
      Component comp = (Component)comps.next();
      Iterator events = comp.getType().eventIterator();
      while (events.hasNext()) {
	Event event = (Event)events.next();
	if (comp.isSpontaneousEvent(event)) {
	  Port port = event.getPort();
	  Message msg = event.getMessage();
	  if (debug) {
	    System.err.println("spontaneous: " + comp.getName()
			       + " " + port.getName()
			       + " " + msg.getLabel());
	  }
	  HashMap current = new HashMap();
	  Set cset = findCascades(sys, ccrules, comp, port, msg, current);
	  if (debug) {
	    System.err.println("found " + cset.size() + " cascades");
	  }
	  Iterator acts = cset.iterator();
	  int aindex = 1;
	  while (acts.hasNext()) {
	    HashMap act = (HashMap)acts.next();
	    writePropositionalAction(sys, comp, port, msg, aindex++, act, out);
	  }
	}
      }
    }

    // observation update action
    out.println("  (:action advance");
    out.println("    :parameters (?o)");
    out.println("    :precondition (and (future ?o)");
    out.println("                       (forall (?op) (imply (precedes ?op ?o) (observed ?op))))");
    out.println("    :effect (and (not (future ?o)) (pending ?o)))");

    // end of domain definition
    out.println(" )");

    // system initial state as problem (should be "situation")
    out.println("(define (problem INIT)");
    out.println("  (:domain SYSTEM)");
    out.print("  (:init");
    Configuration conf = sys.getInitialConfiguration();
    comps = sys.componentIterator();
    while (comps.hasNext()) {
      Component comp = (Component)comps.next();
      Iterator vars = comp.variableIterator();
      while (vars.hasNext()) {
	Variable var = (Variable)vars.next();
	Assignement ia = conf.getAssignement(comp, var);
	Value val = ia.getValue();
	out.print(" (state_" + comp.getName()
			 + "_" + var.getName() + "_" + val.getLabel()
			 + ")");
      }
    }
    out.println(")");

    // end initial state description
    out.println(" )");
  }

  private static void writeScenarioPropositional
    (CompositeSystem sys, Map obs, Map po, PrintStream out)
  {
    out.println("(define (problem SCENARIO)");
    out.println("  (:domain SYSTEM)");
    out.print("  (:objects");
    Iterator ids = obs.keySet().iterator();
    while (ids.hasNext()) {
      String id = (String)ids.next();
      out.print(" o_" + id);
    }
    out.println(")");

    out.print("  (:init");
    ids = obs.keySet().iterator();
    while (ids.hasNext()) {
      String id = (String)ids.next();
      ComponentEvent ce = (ComponentEvent)obs.get(id);
      out.print(" (matches_" + ce.getComponent().getName()
		       + "_" + ce.getEvent().getPort().getName()
		       + "_" + ce.getEvent().getMessage().getLabel()
		       + " o_" + id + ")");
      if (po.containsKey(id)) {
	String[] prec = (String[])po.get(id);
	for (int i = 0; i < prec.length; i++)
	  out.print(" (precedes o_" + prec[i] + " o_" + id + ")");
	out.print(" (future o_" + id + ")");
      }
      else {
	out.print(" (pending o_" + id + ")");
      }
    }
    out.println(")");

    if (!obs.keySet().isEmpty()) {
      out.print("  (:goal (and");
      ids = obs.keySet().iterator();
      while (ids.hasNext()) {
	String id = (String)ids.next();
	out.print(" (observed o_" + id + ")");
      }
      out.println("))");
    }

    out.println(" )");
  }

  private static void writeHypothesisPropositional
    (CompositeSystem sys, Set non_faults, PrintStream out)
  {
    out.println("(define (problem HYPOTHESIS)");
    out.println("  (:domain SYSTEM)");
    out.print("  (:goal (and");
    Iterator nfs = non_faults.iterator();
    while (nfs.hasNext()) {
      ComponentEvent ce = (ComponentEvent)nfs.next();
      out.print(" (not (occurred_" + ce.getComponent().getName()
		+ "_" + ce.getEvent().getPort().getName()
		+ "_" + ce.getEvent().getMessage().getLabel() + "))");
    }
    out.println("))");

    out.println(" )");
  }

  private static boolean loadScenario
    (String name, CompositeSystem sys, Map obs, Map po)
    throws IOException
  {
    try {
      FileReader frd = new FileReader(name);
      LineNumberReader lrd = new LineNumberReader(frd);
      String ln = lrd.readLine();
      Pattern p = Pattern.compile("([^ ]+)[ \\t]+([^{]*)\\{([^}]*)\\}");
      while (ln != null) {
	// System.err.println("info: read line \"" + ln + "\"");
	Matcher m = p.matcher(ln);
	if (m.matches()) {
	  if (m.groupCount() < 3) {
	    System.err.println("error: line at " + name + ":"
			       + Integer.toString(lrd.getLineNumber())
			       + " \"" + ln + "\" matches but group count = "
			       + Integer.toString(m.groupCount()));
	    System.exit(255);
	  }
	  // System.err.println("info: matched [" + m.group(0) + "]");
	  String id = m.group(1);
	  String desc = m.group(2);
	  String prec = m.group(3).trim();
	  ComponentEvent ce =
	    ComponentEvent.read(desc, sys.getInitialTopology());
	  if (ce == null) {
	    System.err.println("warning: event description \"" + desc
			       + "\" could not be interpreted");
	  }
	  else {
	    obs.put(id, ce);
	    if (prec.length() > 0) {
	      String[] precs = prec.split("\\s+");
	      po.put(id, precs);
	    }
	  }
	}
	else {
	  System.err.println("warning: line \"" + ln + "\" at " + name + ":"
			     + Integer.toString(lrd.getLineNumber())
			     + " not matched");
	}
	ln = lrd.readLine();
      }
      frd.close();
    }
    catch (java.io.FileNotFoundException fnf) {
      System.err.println("error: file " + name + " not found");
      return false;
    }
    return true;
  }

  private static boolean loadComponentEventSet
    (String name, CompositeSystem sys, Set set)
    throws IOException
  {
    try {
      FileReader frd = new FileReader(name);
      LineNumberReader lrd = new LineNumberReader(frd);
      String ln = lrd.readLine();
      while (ln != null) {
	ComponentEvent ce =
	  ComponentEvent.read(ln, sys.getInitialTopology());
	if (ce == null) {
	  System.err.println("warning: event description \"" + ln
			     + "\" could not be interpreted");
	}
	else {
	  set.add(ce);
	}
	ln = lrd.readLine();
      }
      frd.close();
    }
    catch (java.io.FileNotFoundException fnf) {
      System.err.println("error: file " + name + " not found");
      return false;
    }
    return true;
  }

  private static void makeCompletionRules
    (Event event, Object[] rules, int p, Map cond, Collection newrules)
  {
    if (p < rules.length) {
      Rule next_rule = (Rule)rules[p];
      HashMap m = new HashMap();
      getConjuncts(next_rule.getPrecondition(), m, 0);
      Iterator i = m.entrySet().iterator();
      while (i.hasNext()) {
	Map.Entry e = (Map.Entry)i.next();
	Variable var = (Variable)e.getKey();
	VariableCondition vc = (VariableCondition)e.getValue();
	if (cond.containsKey(var)) {
	  // if the condition under construction already has a condition
	  // on var, check if the current condition is consistent with vc:
	  // if it's not, the current rule is already prevented; else, if
	  // the current condition implies vc, the current rule cannot be
	  // prevented using this variable
	  VariableCondition cvc = (VariableCondition)cond.get(var);
	  if (!cvc.consistentWith(vc, var)) {
	    makeCompletionRules(event, rules, p + 1, cond, newrules);
	  }
	  else if (!cvc.implies(vc, var)) {
	    // the current condition on var neither contradicts nor
	    // implies vc: this can only happen if the current cond is
	    // a not-eq.
	    assert(cvc.isNotEq());
	    if (vc.isEq()) {
	      // replace previous condition (=/=) with new (==)
	      cond.put(var, vc);
	      makeCompletionRules(event, rules, p + 1, cond, newrules);
	    }
	    else {
	      // we have two not-eq's on var: replace with a single
	      // eq for each other value
	      Iterator domvals = var.getDomain().valueIterator();
	      while (domvals.hasNext()) {
		Value dv = (Value)domvals.next();
		if (!dv.equals(cvc.getValue()) && !dv.equals(vc.getValue())) {
		  cond.put(var, new VariableCondition(dv));
		  makeCompletionRules(event, rules, p + 1, cond, newrules);
		}
	      }
	    }
	    // restore previous condition after we're done
	    cond.put(var, cvc);
	  }
	}
	else {
	  // if the current condition does not assign 'var', add a
	  // val=val or var=/=val condition that is the opposite of
	  // the vc
	  cond.put(var, vc.getNegation());
	  makeCompletionRules(event, rules, p + 1, cond, newrules);
	  // remove 'var' from cond after we're done
	  cond.remove(var);
	}
      }
    }
    else {
      // done: cond now holds a set of var/val pairs that ensure no
      // rule in rules is applicable; create a new rule with this as
      // precondition, event as trigger, and no output or effect.
      Rule r = new Rule(makeCondition(cond), event,
			new Vector(), new RuleEffect(new HashMap()));
      newrules.add(r);
    }
  }

  private static Collection completion(ComponentType ct)
  {
    Vector newrules = new Vector();
    Iterator events = ct.eventIterator();
    while (events.hasNext()) {
      Event event = (Event)events.next();
      // only process events on non-spontaneous input ports
      if ((ct.getInputPort(event.getPort().getName()) != null) &&
	  !ct.isSpontaneousEvent(event)) {
	Object[] rules = ct.rulesOfEvent(event).toArray();
	HashMap cond = new HashMap();
	makeCompletionRules(event, rules, 0, cond, newrules);
      }
    }
    return newrules;
  }

  private static Map completion(CompositeSystem sys)
  {
    HashMap m = new HashMap();
    Iterator cts = sys.typeIterator();
    while (cts.hasNext()) {
      ComponentType ct = (ComponentType)cts.next();
      m.put(ct, completion(ct));
    }
    return m;
  }

  public static void main(String[] argv) {
    boolean propositional = true;
    boolean do_completion = false;
    boolean have_scenario = false;
    boolean do_test = false;
    int     max_faults = 0;
    String  system_file = null;
    String  scenario_file = null;
    String  faults_file = null;
    String  file_name = null;
    for (int k = 0; k < argv.length; k++) {
      if (argv[k].equalsIgnoreCase("-s") && (k < argv.length - 1)) {
	system_file = argv[k + 1];
	k += 1;
      }
      else if (argv[k].equalsIgnoreCase("-o") && (k < argv.length - 1)) {
	scenario_file = argv[k + 1];
	k += 1;
      }
      else if (argv[k].equalsIgnoreCase("-f") && (k < argv.length - 1)) {
	faults_file = argv[k + 1];
	k += 1;
      }
      else if (argv[k].equalsIgnoreCase("-m") && (k < argv.length - 1)) {
	max_faults = Integer.parseInt(argv[k + 1]);
	k += 1;
      }
      else if (argv[k].equalsIgnoreCase("-n") && (k < argv.length - 1)) {
	file_name = argv[k + 1];
	k += 1;
      }
      else if (argv[k].equalsIgnoreCase("-c"))
	strict_non_cyclic = false;
      else if (argv[k].equalsIgnoreCase("-i"))
	do_completion = true;
      else if (argv[k].equalsIgnoreCase("-v"))
	write_comments = true;
      else if (argv[k].equalsIgnoreCase("-d"))
	debug = true;
      else if (argv[k].equalsIgnoreCase("-p"))
	propositional = false;
      else if (argv[k].equalsIgnoreCase("-t"))
	do_test = true;
      else if (!argv[k].startsWith("-") && (system_file == null))
	system_file = argv[k];
    }
    if (system_file == null) {
      System.err.println("csdltopddl [-c] [-i] [-v] <system file> [-o <scenario file>] [-f <faults file>] [-m <max faults>] [-n <output file name>]");
      System.exit(1);
    }
    try {
      System.err.println("loading system " + system_file + "...");
      CompositeSystem sys = CompositeSystem.load(system_file);
      Map ccrules = (do_completion ? completion(sys) : new HashMap());
      HashMap obs = null;
      HashMap po = null;
      if (scenario_file != null) {
	obs = new HashMap();
	po = new HashMap();
	System.err.println("loading scenario " + scenario_file + "...");
	have_scenario = loadScenario(scenario_file, sys, obs, po);
      }
      HashSet faults = new HashSet();
      if (faults_file != null) {
	System.err.println("loading fault event set " + faults_file + "...");
	loadComponentEventSet(faults_file, sys, faults);
      }
      if (do_test) {
	if (faults.isEmpty()) {
	  System.err.println("can't do enumerator test without a set");
	  System.exit(1);
	}
	SubsetEnumerator e = new SubsetEnumerator(faults);
	for (int n = 0; n <= max_faults; n++) {
	  System.err.println("all sets of " + Integer.toString(n) + " faults:");
	  boolean more = e.first(faults.size() - n);
	  while (more) {
	    Iterator i = e.iterator();
	    System.err.print("{");
	    while (i.hasNext()) {
	      ComponentEvent ce = (ComponentEvent)i.next();
	      System.err.print(" " + ce.getComponent().getName()
			       + "_" + ce.getEvent().getPort().getName()
			       + "_" + ce.getEvent().getMessage().getLabel());
	    }
	    System.err.println("}");
	    more = e.next(faults.size() - n);
	  }
	}
      }
      else if (propositional) {
	if (file_name != null) {
	  FileOutputStream fs =
	    new FileOutputStream(file_name + ".pddl");
	  PrintStream ps = new PrintStream(fs);
	  writeSystemPropositional(sys, ccrules, ps);
	  if (have_scenario)
	    writeScenarioPropositional(sys, obs, po, ps);
	  ps.close();
	}
	else {
	  writeSystemPropositional(sys, ccrules, System.out);
	  if (have_scenario)
	    writeScenarioPropositional(sys, obs, po, System.out);
	}
	if ((max_faults > 0) && !faults.isEmpty()) {
	  for (int n = 0; n <= max_faults; n++) {
	    System.err.println("writing " + Integer.toString(n)
			       + "-fault hypotheses...");
	    SubsetEnumerator e = new SubsetEnumerator(faults);
	    boolean more = e.first(faults.size() - n);
	    int i = 1;
	    while (more) {
	      FileOutputStream fs =
		new FileOutputStream(file_name + "-" + Integer.toString(n)
				     + "-faults-" + i + ".pddl");
	      i += 1;
	      PrintStream ps = new PrintStream(fs);
	      writeHypothesisPropositional(sys, e, ps);
	      ps.close();
	      more = e.next(faults.size() - n);
	    }
	  }
	}
      }
      else {
	writeDomain(sys, System.out);
      }
    }
    catch (Exception e) {
      System.err.println("caught " + e.getMessage());
      e.printStackTrace();
    }
  }

}
